==> GFS-4.1.0.tar.gz: GFS-4.1.0/src/stomith/daemon/stomithd.c <==
/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
**
**    Module name:
**      stomithd.c
**
**    Description:  Selects a specific stomith agent program to execute based
**                  on data from kernel stomith module.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <syslog.h>

#include "stomith.h"

#define PIDFILE "/tmp/.stomithd.pid"


void inet_ntoa(uint32 ip, char *buf)
{
  int i;
  char *p;

  p = buf;

  for (i = 3; i >= 0; i--)
  {
    p += sprintf(p, "%d", (ip >> (8 * i)) & 0xFF);
    *(p++) = '.';
  }
}


/*
 * Startup code to make a proper daemon.  Much from Richard Stevens APUE.
 */

void daemon_startup(void)
{
  pid_t pid;
  struct flock lock;
  char buf[10];
  int fd, val, i;


  if ((pid = fork()) < 0)
  {
    perror("can't fork new process");
    exit(-1);
  }
  else if (pid != 0)
  {
    exit(0);
  }

  for (i = 0; i < 10; i++)
    close(i);


  setsid();	/* become session leader */
  chdir("/");	/* change working directory */
  umask(0);	/* clear file mode creation masks */

  openlog("stomithd", 0, LOG_DAEMON);


  /* prevent multiple stomith daemons from running at once 
     by locking a file */

  if ((fd = open(PIDFILE, O_WRONLY | O_CREAT, 
		 (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))) < 0)
  {
    fprintf(stderr, "open error on pidfile - %d\n", errno);
    exit(-1);
  }

  lock.l_type = F_WRLCK;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;

  if (fcntl(fd, F_SETLK, &lock) < 0)
  {
    if (errno == EACCES || errno == EAGAIN)
    {
      fprintf(stderr, "stomithd already running\n");
      exit(0);
    }
    else
    {
      fprintf(stderr, "fcntl F_SETLK error - %d\n", errno);
      exit(-1);
    }
  }

  if (ftruncate(fd, 0) < 0)
  {
    fprintf(stderr, "ftruncate error - %d\n", errno);
    exit(-1);
  }

  sprintf(buf, "%d\n", getpid());
  if (write(fd, buf, strlen(buf)) != strlen(buf))
  {
    fprintf(stderr, "write error to \"%s\" - %d\n", PIDFILE, errno);
    exit(-1);
  }

  if ((val = fcntl(fd, F_GETFD, 0)) < 0)
  {
    fprintf(stderr, "fcntl F_GETFD error - %d\n", errno);
    exit(-1);
  }

  val |= FD_CLOEXEC;
  if (fcntl(fd, F_SETFD, val) < 0)
  {
    fprintf(stderr, "fcntl F_SETFD error - %d\n", errno);
    exit(-1);
  }
}



int main(int argc, char **argv)
{
  cluster_stomith_t *victim, *local;
  int fd, len, error, success, status = 0, val = 0;
  unsigned int minor;
  char buf[2 * sizeof(cluster_stomith_t)];
  char line[256], result[256], desc[256], name[32];
  char args[256], cmd[32];
  unsigned char c = 0xFF;
  int fds[2], fds2[2];
  FILE *pf;



  /* set up the char device node */

  pf = fopen("/proc/misc", "r");
  if (!pf)
  {
    fprintf(stderr, "can't open /proc/misc - %d\n", errno);
    exit(-1);
  }

  while ((error = fscanf(pf, "%u %s\n", &minor, name)) != EOF)   
    if (!strncmp(name, "stomith", 7))
      break;

  if (error == EOF)
  {
    fprintf(stderr, "No stomith found in /proc/misc\n");
    exit(-1);
  }
  
  fclose(pf);

  unlink("/dev/stomith");
  error = mknod("/dev/stomith", 0600 | S_IFCHR, makedev(10, minor));
  if (error < 0)
  {
    fprintf(stderr, "can't make node /dev/stomith - %d\n", errno);
    exit(-1);
  }


  /* stomith requests come from /dev/stomith */

  fd = open("/dev/stomith", O_RDWR);
  if (fd < 0)
  {
    fprintf(stderr, "can't open /dev/stomith - %d\n", errno);
    exit(-1);
  }


  daemon_startup();


  fd = open("/dev/stomith", O_RDWR);
  if (fd < 0)
  { 
    syslog(LOG_CRIT, "can't open /dev/stomith - %d\n", errno);
    exit(-1);
  } 

  
  /* make any outstanding stomith requests fail */

  if (write(fd, &c, 1) != 1)
  {
    syslog(LOG_CRIT, "can't write to /dev/stomith - %d\n", errno);
    exit(-1);
  }


  len = sizeof(cluster_stomith_t);
  victim = (cluster_stomith_t *)buf;
  local = (cluster_stomith_t *)(buf + len);

  
  /* read stomith requests from kernel stomith module */  

  while (1)
  {
    memset(cmd, 0, 32);
    memset(args, 0, 256);

    error = lseek(fd, 0, SEEK_SET);
    if (error)
    {
      syslog(LOG_CRIT, "can't seek on /dev/stomith - %d\n", errno);
      exit(-1);
    }

    error = read(fd, &buf, len * 2);
    if (error < 0)
    {
      syslog(LOG_CRIT, "can't read /dev/stomith - %d\n", errno);
      exit(-1);
    }

    switch (victim->cs_magic)
    {
    case STM_VACM_EMP:
    {
      stm_vacm_emp_t *va = (stm_vacm_emp_t *)(&victim->cs_st);
      sprintf(cmd, "do_vacm_emp");           
      sprintf(args, "-x %s -l %s -p %s -n %s\n", 
	      va->st_nexxus, 
	      va->st_login,
	      va->st_passwd, 
	      va->st_nodeid);
      break;
    }      

    case STM_WTI_NPS:
    {
      stm_wti_nps_t *nps = (stm_wti_nps_t *)(&victim->cs_st);
      sprintf(cmd, "do_wti_nps");
      sprintf(args, "-a %u -n %u -p %s\n", 
	      nps->st_ipaddr, 
	      nps->st_plug,
	      nps->st_passwd);
      break;
    }
	
    case STM_BRCD_ZONE:
    {
      stm_brcd_zone_t *bz = (stm_brcd_zone_t *)(&victim->cs_st);
      sprintf(cmd, "do_brcd_zone");
      sprintf(args, "-a %u -l %s -p %s -c %s -f %s -z %s -n %s\n",
	      bz->st_ipaddr,
	      bz->st_login,
	      bz->st_passwd, 
	      bz->st_command, 
	      bz->st_cfgname,
	      bz->st_comzone, 
	      bz->st_actname);
      break;
    }

    case STM_BRCD_PORT:
    {
      stm_brcd_port_t *bp = (stm_brcd_port_t *)(&victim->cs_st);
      sprintf(cmd, "do_brcd_port");
      sprintf(args, "-a %u -l %s -p %s -n %d\n",
	      bp->st_ipaddr,
	      bp->st_login,
	      bp->st_passwd, 
	      bp->st_port);
      break;
    }

    case STM_GNBD:
    {
      stm_gnbd_t *sg = (stm_gnbd_t *)(&victim->cs_st);
      sprintf(cmd, "do_gclient_stm");
      sprintf(args, "-s %u\n",
	      sg->st_ipaddr);
      break;
    }
       
    case STM_MEATWARE:
    case STM_MANUAL:
    {
      stm_meatware_t *meat = (stm_meatware_t *)(&victim->cs_st);
      char ip[16];
      sprintf(cmd, "wait_meatware");
      inet_ntoa(meat->st_ipaddr, ip);
      sprintf(args, "-s %s\n", ip);
      break;
    }

    case STM_VXHUB_BYPASS:
    {
      stm_vx_hub_t *hub = (stm_vx_hub_t *)(&victim->cs_st);
      sprintf(cmd, "do_vx_hub");
      sprintf(args, "-a %u -n %u -p %s\n", 
              hub->st_ipaddr, 
              hub->st_port,
              hub->st_passwd);
      break;
    }   

	
    default:
      syslog(LOG_CRIT, "invalid stomith method (%u)\n", victim->cs_magic);
      continue;
    }       


    if (pipe(fds))
    {
      syslog(LOG_CRIT, "can't create pipe - %d\n", errno);
      error = -1;
      
      if (write(fd, &error, sizeof(error) != sizeof(error)))
	syslog(LOG_CRIT, "bad write to /proc/stomith - %d\n", errno);

      continue;
    }
    
    if (pipe(fds2))
    {
      syslog(LOG_CRIT, "can't create pipe - %d\n", errno);
      error = -1;
      
      if (write(fd, &error, sizeof(error) != sizeof(error)))
	syslog(LOG_CRIT, "bad write to /proc/stomith - %d\n", errno);
      
      close(fds[0]);
      close(fds[1]);

      continue;
    }
      
    
              
    if (!fork())
    {
      /* set output to go back to stomithd */
      close(1);
      dup(fds[1]);
      
      /* set input to come from stomithd */
      close(0);
      dup(fds2[0]);

      /* close unneeded file descriptors */
      close(fds[0]);
      close(fds[1]);
      close(fds2[0]);
      close(fds2[1]);
      close(3);

      execlp(cmd, cmd, NULL);
      
      printf("STOMITH:failed:failed to exec agent (%s), errno: %d\n",
	     cmd, errno);
      return(0);
    }

    else
    {
      /* set input to come from agent */
      close(0);
      dup(fds[0]);

      success = 0;

      /* set fd to nonblocking in case agent dies with no output */
      if ((val = fcntl(0, F_GETFL, 0)) < 0)
	syslog(LOG_CRIT, "fcntl can't get flags\n");

      val |= O_NONBLOCK;
      
      if (fcntl(0, F_SETFL, val) < 0)
	syslog(LOG_CRIT, "fcntl can't set flags\n");

      /* send stomith parameters to agent */
      if (write(fds2[1], args, sizeof(args)) != sizeof(args))
      {
	sprintf(desc, "write args to agent failed\n");
	goto end;
      }

      wait(&status);

      memset(line, 0, 256);
      memset(result, 0, 256);
      memset(desc, 0, 256);

      /* agent should return a result */
      error = read(0, line, 256);      
      if (error < 0)
      {
	syslog(LOG_CRIT, "read from agent returned %d\n", error);
	error = 0;
      }

      line[error] = '\0';      
      sscanf(line, "STOMITH:%[^:]:%[^\n]", result, desc);
      
      if (!strncmp(result, "success", 7))
	success = 1;
      
    end:
      if (success)
      {
	error = 21;
	syslog(LOG_CRIT, "successful stomith - %s\n", desc);
      }
      else
      {
	error = -1;
	syslog(LOG_CRIT, "failed stomith - %s\n", line);
      }

      if (write(fd, &error, sizeof(error)) != sizeof(error))
	syslog(LOG_CRIT, "bad write to /dev/stomith - %d\n", errno);

      close(fds2[0]);
      close(fds2[1]);
      close(fds[0]);
      close(fds[1]);
    }
  }

  close(fd);

  return(0);
}

==> GFS-4.1.1.tar.gz: GFS-4.1.1/src/stomith/daemon/stomithd.c <==
/******************************************************************************
*******************************************************************************
**
**    Copyright 2000-2001 Sistina Software, Inc.
**
**    This is free software released under the GNU General Public License.
**    There is no warranty for this software.  See the file COPYING for
**    details.
**
**    See the file CONTRIBUTORS for a list of contributors.
**
**    This file is maintained by:
**      David C. Teigland <teigland@sistina.com>
**
**    Module name:
**      stomithd.c
**
**    Description:  Selects a specific stomith agent program to execute based
**                  on data from kernel stomith module.
**
*******************************************************************************
******************************************************************************/

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fcntl.h>
#include <errno.h>
#include <ctype.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/stat.h>
#include <sys/sysmacros.h>
#include <syslog.h>

#include "stomith.h"

#define PIDFILE "/var/run/stomithd.pid"



void inet_ntoa(uint32 ip, char *buf)
{
  int i;
  char *p;

  p = buf;

  for (i = 3; i >= 0; i--)
  {
    p += sprintf(p, "%d", (ip >> (8 * i)) & 0xFF);
    *(p++) = '.';
  }
}


/*
 * Startup code to make a proper daemon.  Much from Richard Stevens APUE.
 */

void daemon_startup(void)
{
  pid_t pid;
  struct flock lock;
  char buf[10];
  int fd, val, i;


  if ((pid = fork()) < 0)
  {
    perror("can't fork new process");
    exit(-1);
  }
  else if (pid != 0)
  {
    exit(0);
  }

  for (i = 0; i < 10; i++)
    close(i);


  setsid();	/* become session leader */
  chdir("/");	/* change working directory */
  umask(0);	/* clear file mode creation masks */

  openlog("stomithd", 0, LOG_DAEMON);


  /* prevent multiple stomith daemons from running at once 
     by locking a file */

  if ((fd = open(PIDFILE, O_WRONLY | O_CREAT, 
		 (S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH))) < 0)
  {
    fprintf(stderr, "open error on pidfile - %d\n", errno);
    exit(-1);
  }

  lock.l_type = F_WRLCK;
  lock.l_start = 0;
  lock.l_whence = SEEK_SET;
  lock.l_len = 0;

  if (fcntl(fd, F_SETLK, &lock) < 0)
  {
    if (errno == EACCES || errno == EAGAIN)
    {
      fprintf(stderr, "stomithd already running\n");
      exit(0);
    }
    else
    {
      fprintf(stderr, "fcntl F_SETLK error - %d\n", errno);
      exit(-1);
    }
  }

  if (ftruncate(fd, 0) < 0)
  {
    fprintf(stderr, "ftruncate error - %d\n", errno);
    exit(-1);
  }

  sprintf(buf, "%d\n", getpid());
  if (write(fd, buf, strlen(buf)) != strlen(buf))
  {
    fprintf(stderr, "write error to \"%s\" - %d\n", PIDFILE, errno);
    exit(-1);
  }

  if ((val = fcntl(fd, F_GETFD, 0)) < 0)
  {
    fprintf(stderr, "fcntl F_GETFD error - %d\n", errno);
    exit(-1);
  }

  val |= FD_CLOEXEC;
  if (fcntl(fd, F_SETFD, val) < 0)
  {
    fprintf(stderr, "fcntl F_SETFD error - %d\n", errno);
    exit(-1);
  }
}



int main(int argc, char **argv)
{
  cluster_stomith_t *victim, *local;
  int fd, len, error, success, status = 0, val = 0;
  unsigned int minor;
  char buf[2 * sizeof(cluster_stomith_t)];
  char line[256], result[256], desc[256], name[32];
  char args[256], cmd[32];
  unsigned char c = 0xFF;
  int fds[2], fds2[2];
  FILE *pf;



  /* set up the char device node */

  pf = fopen("/proc/misc", "r");
  if (!pf)
  {
    fprintf(stderr, "can't open /proc/misc - %d\n", errno);
    exit(-1);
  }

  while ((error = fscanf(pf, "%u %s\n", &minor, name)) != EOF)   
    if (!strncmp(name, "stomith", 7))
      break;

  if (error == EOF)
  {
    fprintf(stderr, "No stomith found in /proc/misc\n");
    exit(-1);
  }
  
  fclose(pf);

  unlink("/dev/stomith");
  error = mknod("/dev/stomith", 0600 | S_IFCHR, makedev(10, minor));
  if (error < 0)
  {
    fprintf(stderr, "can't make node /dev/stomith - %d\n", errno);
    exit(-1);
  }


  /* stomith requests come from /dev/stomith */

  fd = open("/dev/stomith", O_RDWR);
  if (fd < 0)
  {
    fprintf(stderr, "can't open /dev/stomith - %d\n", errno);
    exit(-1);
  }


  daemon_startup();


  fd = open("/dev/stomith", O_RDWR);
  if (fd < 0)
  { 
    syslog(LOG_CRIT, "can't open /dev/stomith - %d\n", errno);
    exit(-1);
  } 

  
  /* make any outstanding stomith requests fail */

  if (write(fd, &c, 1) != 1)
  {
    syslog(LOG_CRIT, "can't write to /dev/stomith - %d\n", errno);
    exit(-1);
  }


  len = sizeof(cluster_stomith_t);
  victim = (cluster_stomith_t *)buf;
  local = (cluster_stomith_t *)(buf + len);

  
  /* read stomith requests from kernel stomith module */  

  while (1)
  {
    memset(cmd, 0, 32);
    memset(args, 0, 256);

    error = lseek(fd, 0, SEEK_SET);
    if (error)
    {
      syslog(LOG_CRIT, "can't seek on /dev/stomith - %d\n", errno);
      exit(-1);
    }

    error = read(fd, &buf, len * 2);
    if (error < 0)
    {
      syslog(LOG_CRIT, "can't read /dev/stomith - %d\n", errno);
      exit(-1);
    }

    switch (victim->cs_magic)
    {
    case STM_VACM_EMP:
    {
      stm_vacm_emp_t *va = (stm_vacm_emp_t *)(&victim->cs_st);
      sprintf(cmd, "do_vacm_emp");           
      sprintf(args, "-x %s -l %s -p %s -n %s\n", 
	      va->st_nexxus, 
	      va->st_login,
	      va->st_passwd, 
	      va->st_nodeid);
      break;
    }      

    case STM_WTI_NPS:
    {
      stm_wti_nps_t *nps = (stm_wti_nps_t *)(&victim->cs_st);
      sprintf(cmd, "do_wti_nps");
      sprintf(args, "-a %u -n %u -p %s\n", 
	      nps->st_ipaddr, 
	      nps->st_plug,
	      nps->st_passwd);
      break;
    }
	
    case STM_BRCD_ZONE:
    {
      stm_brcd_zone_t *bz = (stm_brcd_zone_t *)(&victim->cs_st);
      sprintf(cmd, "do_brcd_zone");
      sprintf(args, "-a %u -l %s -p %s -c %s -f %s -z %s -n %s\n",
	      bz->st_ipaddr,
	      bz->st_login,
	      bz->st_passwd, 
	      bz->st_command, 
	      bz->st_cfgname,
	      bz->st_comzone, 
	      bz->st_actname);
      break;
    }

    case STM_BRCD_PORT:
    {
      stm_brcd_port_t *bp = (stm_brcd_port_t *)(&victim->cs_st);
      sprintf(cmd, "do_brcd_port");
      sprintf(args, "-a %u -l %s -p %s -n %d\n",
	      bp->st_ipaddr,
	      bp->st_login,
	      bp->st_passwd, 
	      bp->st_port);
      break;
    }

    case STM_GNBD:
    {
      stm_gnbd_t *sg = (stm_gnbd_t *)(&victim->cs_st);
      sprintf(cmd, "do_gclient_stm");
      sprintf(args, "-s %u\n",
	      sg->st_ipaddr);
      break;
    }
       
    case STM_MEATWARE:
    case STM_MANUAL:
    {
      stm_meatware_t *meat = (stm_meatware_t *)(&victim->cs_st);
      char ip[16];
      sprintf(cmd, "wait_meatware");
      inet_ntoa(meat->st_ipaddr, ip);
      sprintf(args, "-s %s\n", ip);
      break;
    }

    case STM_VXHUB_BYPASS:
    {
      stm_vx_hub_t *hub = (stm_vx_hub_t *)(&victim->cs_st);
      sprintf(cmd, "do_vx_hub");
      sprintf(args, "-a %u -n %u -p %s\n", 
              hub->st_ipaddr, 
              hub->st_port,
              hub->st_passwd);
      break;
    }   

    case STM_APC_MS:
    {
      stm_apc_ms_t *apc = (stm_apc_ms_t *)(&victim->cs_st);
      sprintf(cmd, "do_apc_ms");
      sprintf(args, "-a %u -l %s -p %s -n %d\n",
	      apc->st_ipaddr,
	      apc->st_login,
	      apc->st_passwd,
              apc->st_outlet);
      break;
    }
	
    default:
      syslog(LOG_CRIT, "invalid stomith method (%u)\n", victim->cs_magic);
      continue;
    }       


    if (pipe(fds))
    {
      syslog(LOG_CRIT, "can't create pipe - %d\n", errno);
      error = -1;
      
      if (write(fd, &error, sizeof(error) != sizeof(error)))
	syslog(LOG_CRIT, "bad write to /proc/stomith - %d\n", errno);

      continue;
    }
    
    if (pipe(fds2))
    {
      syslog(LOG_CRIT, "can't create pipe - %d\n", errno);
      error = -1;
      
      if (write(fd, &error, sizeof(error) != sizeof(error)))
	syslog(LOG_CRIT, "bad write to /proc/stomith - %d\n", errno);
      
      close(fds[0]);
      close(fds[1]);

      continue;
    }
      
    
              
    if (!fork())
    {
      /* set output to go back to stomithd */
      close(1);
      dup(fds[1]);
      
      /* set input to come from stomithd */
      close(0);
      dup(fds2[0]);

      /* close unneeded file descriptors */
      close(fds[0]);
      close(fds[1]);
      close(fds2[0]);
      close(fds2[1]);
      close(3);

      execlp(cmd, cmd, NULL);
      
      printf("STOMITH:failed:failed to exec agent (%s), errno: %d\n",
	     cmd, errno);
      return(0);
    }

    else
    {
      /* set input to come from agent */
      close(0);
      dup(fds[0]);

      success = 0;

      /* set fd to nonblocking in case agent dies with no output */
      if ((val = fcntl(0, F_GETFL, 0)) < 0)
	syslog(LOG_CRIT, "fcntl can't get flags\n");

      val |= O_NONBLOCK;
      
      if (fcntl(0, F_SETFL, val) < 0)
	syslog(LOG_CRIT, "fcntl can't set flags\n");

      /* send stomith parameters to agent */
      if (write(fds2[1], args, sizeof(args)) != sizeof(args))
      {
	sprintf(desc, "write args to agent failed\n");
	goto end;
      }

      wait(&status);

      memset(line, 0, 256);
      memset(result, 0, 256);
      memset(desc, 0, 256);

      /* agent should return a result */
      error = read(0, line, 256);      
      if (error < 0)
      {
	syslog(LOG_CRIT, "read from agent returned %d\n", error);
	error = 0;
      }

      line[error] = '\0';      
      sscanf(line, "STOMITH:%[^:]:%[^\n]", result, desc);
      
      if (!strncmp(result, "success", 7))
	success = 1;
      
    end:
      if (success)
      {
	error = 21;
	syslog(LOG_CRIT, "successful stomith - %s\n", desc);
      }
      else
      {
	error = -1;
	syslog(LOG_CRIT, "failed stomith - %s\n", line);
      }

      if (write(fd, &error, sizeof(error)) != sizeof(error))
	syslog(LOG_CRIT, "bad write to /dev/stomith - %d\n", errno);

      close(fds2[0]);
      close(fds2[1]);
      close(fds[0]);
      close(fds[1]);
    }
  }

  close(fd);

  return(0);
}
